import types

from django.db.models import QuerySet, Manager
from rest_framework import serializers
from rest_framework.relations import RelatedField

from guardian_queryset.utils import has_view_permissions
from rest_framework_guardian.serializers import ObjectPermissionsAssignmentMixin


def get_queryset_guardian(self):
    queryset = self.queryset
    if isinstance(queryset, (QuerySet, Manager)):
        queryset = queryset.all()
        try:
            queryset = queryset.can_view()
        except:
            pass
        # if hasattr(queryset, "can_view"):
    return queryset


def run_validator_with_class(cls):
    def run_validators(self, value):
        for validator in self.validators:
            try:
                validator.queryset = validator.queryset.can_view()
            except:
                pass
        return super(cls, self).run_validators(value)
    return run_validators


class ValidateForeignKeyUserMixin(serializers.Serializer):

    @property
    def _writable_fields(self):
        for field in self.fields.values():
            if not field.read_only:
                # if isinstance(field, RelatedField) or issubclass(type(field), RelatedField):
                if isinstance(field, RelatedField):
                    field.get_queryset = types.MethodType(get_queryset_guardian, field)
                    field.run_validators = types.MethodType(run_validator_with_class(type(field)), field)
                yield field

    @property
    def _readable_fields(self):
        for field in self.fields.values():
            if not field.write_only:
                if isinstance(field, RelatedField):
                    field.get_queryset = types.MethodType(get_queryset_guardian, field)
                    field.run_validators = types.MethodType(run_validator_with_class(type(field)), field)
                yield field

    def get_unchecked_foreignkey_fields(self):
        return []

    def get_user_field_name(self):
        return 'user'

    # TODO: fix user check
    def validate(self, data):
        current_user = self.context['request'].user
        for name, field in self.get_fields().items():
            if isinstance(field, RelatedField) or issubclass(type(field), RelatedField):
            # if type(field) == PrimaryKeyRelatedField:
                if name not in self.get_unchecked_foreignkey_fields():
                    if not field.read_only:
                        if name == self.get_user_field_name():
                            if data[self.get_user_field_name()] != current_user:
                                raise serializers.ValidationError("attempted to create foreign key on unauthorized object")
                        elif name in data and not has_view_permissions(current_user, data):
                            raise serializers.ValidationError("attempted to create foreign key on unauthorized object")
        return data

class AssignActiveUserDefaultPermissionMixin(ObjectPermissionsAssignmentMixin):

    def get_permissions_map(self, created):
        current_user = self.context['request'].user
        model_name = self.Meta.model.__name__.lower()

        return {
            'view_' + model_name: [current_user],
            'change_' + model_name: [current_user],
            'delete_' + model_name: [current_user]
        }




from django.db.models.manager import BaseManager
from guardian_queryset.queryset import GuardianQuerySet
from django.db import models

class GuardianManager(BaseManager.from_queryset(GuardianQuerySet), models.Manager):
    pass





from django.conf import settings
from django.db.models import ForeignKey
from guardian.shortcuts import assign_perm, remove_perm
from rest_framework.exceptions import ValidationError

PUBLIC_VIEW_MODELS_LOWER = [model_name.lower() for model_name in getattr(settings, 'PUBLIC_VIEW_MODELS', [])]

def get_model_name(obj_or_query):
    try:
        return obj_or_query.model.__name__.lower()
    except Exception:
        return obj_or_query._meta.model.__name__.lower()

def assign_default_permissions(user, object):
    model_name = get_model_name(object)
    assign_perm('view_' + model_name, user, object)
    assign_perm('change_' + model_name, user, object)
    assign_perm('delete_' + model_name, user, object)


def assign_view_permissions(user, object, save=True):
    model_name = get_model_name(object)
    assign_perm('view_' + model_name, user, object)


def assign_change_permissions(user, object, save=True):
    model_name = get_model_name(object)
    assign_perm('change_' + model_name, user, object)


def remove_view_permissions(user, object):
    model_name = get_model_name(object)
    remove_perm('view_' + model_name, user, object)


def has_permissions(user, object, permission):
    model_name = get_model_name(object)
    return user.has_perm(permission + model_name, object)


def has_view_permissions(user, object):
    model_name = get_model_name(object)
    return model_name in PUBLIC_VIEW_MODELS_LOWER or has_permissions(user, object, 'view_')

def has_change_permissions(user, object):
    return has_permissions(user, object, 'change_')

def has_delete_permissions(user, object):
    return has_permissions(user, object, 'delete_')


def validate_foreign_keys(user, object, user_field_name="user", unchecked_field_names=[]):
    for attr_name in dir(object._meta.model):
        attr = getattr(object._meta.model, attr_name)
        if hasattr(attr, 'field'):
            field = attr.field
            if isinstance(attr.field, ForeignKey):
                if hasattr(object, field.name) and field.name not in unchecked_field_names:
                    related_object = getattr(object, field.name)
                    if related_object:
                        from guardian_queryset.models import GuardianModel
                        if field.name == user_field_name:
                            if related_object != user:
                                raise ValidationError(f"attempted to create foreign key on unauthorized object {related_object}")
                        elif isinstance(related_object, GuardianModel) and not has_view_permissions(user, related_object):
                            raise ValidationError(f"attempted to create foreign key on unauthorized object {related_object}")


from django.db import models

from global_requests import get_thread_user
from guardian_queryset.manager import GuardianManager
from guardian_queryset.queryset import InvalidOperation
from guardian_queryset.utils import assign_default_permissions, has_change_permissions, has_delete_permissions
from django_utils.utils import is_anonymous_user


class GuardianModel(models.Model):

    # __default_manager = _InternalGuardianManager()
    objects = GuardianManager()
    _objects = models.Manager()

    # @classmethod
    # def objects(cls, user):
    #     return cls._guardian_objects.with_user(user)

    class Meta:
        abstract = True

    def get_user(self, user, use_obj_user):
        user = user or get_thread_user()
        if use_obj_user and is_anonymous_user(user):
            try:
                user = self.user
            except:
                raise AssertionError("Model doesn't have a user")
        assert not user.is_anonymous, "User for saving is anonymous"
        return user

    def _save(self, *args, **kwargs):
        return super().save(*args, **kwargs)

    def save(self, *args, user=None, use_obj_user=True, **kwargs):
        user = self.get_user(user, use_obj_user)
        if self._state.adding or not self.pk:
            super().save(*args, **kwargs)
            assign_default_permissions(user, self)
            return super().save()
        else:
            if not has_change_permissions(user, self):
                raise InvalidOperation("User does not have permission to update this object")
            return super().save(*args, **kwargs)


    def _delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    def delete(self, *args, user=None, use_obj_user=True, **kwargs):
        user = self.get_user(user, use_obj_user)
        if not has_delete_permissions(user, self):
            raise InvalidOperation("User does not have permission to delete this object")
        return super().delete(*args, **kwargs)

    # def _perform_unique_checks(self, unique_checks):
    #     errors = {}
    #
    #     for model_class, unique_check in unique_checks:
    #         # Try to look up an existing object with the same values as this
    #         # object's values for all the unique field.
    #
    #         lookup_kwargs = {}
    #         for field_name in unique_check:
    #             f = self._meta.get_field(field_name)
    #             lookup_value = getattr(self, f.attname)
    #             # TODO: Handle multiple backends with different feature flags.
    #             if (lookup_value is None or
    #                     (lookup_value == '' and connection.features.interprets_empty_strings_as_nulls)):
    #                 # no value, skip the lookup
    #                 continue
    #             if f.primary_key and not self._state.adding:
    #                 # no need to check for unique primary key when editing
    #                 continue
    #             lookup_kwargs[str(field_name)] = lookup_value
    #
    #         # some fields were skipped, no reason to do the check
    #         if len(unique_check) != len(lookup_kwargs):
    #             continue
    #
    #
    #         qs = model_class._default_manager.filter(**lookup_kwargs)
    #         try:
    #             qs = qs.force_validate()
    #         except Exception:
    #             pass
    #
    #         # Exclude the current object from the query if we are editing an
    #         # instance (as opposed to creating a new one)
    #         # Note that we need to use the pk as defined by model_class, not
    #         # self.pk. These can be different fields because model inheritance
    #         # allows single model to have effectively multiple primary keys.
    #         # Refs #17615.
    #         model_class_pk = self._get_pk_val(model_class._meta)
    #         if not self._state.adding and model_class_pk is not None:
    #             qs = qs.exclude(pk=model_class_pk)
    #         if qs.exists():
    #             if len(unique_check) == 1:
    #                 key = unique_check[0]
    #             else:
    #                 key = NON_FIELD_ERRORS
    #             errors.setdefault(key, []).append(self.unique_error_message(model_class, unique_check))
    #
    #     return errors
    #
    # def clean_fields(self, exclude=None):
    #     """
    #     Clean all fields and raise a ValidationError containing a dict
    #     of all validation errors if any occur.
    #     """
    #     if exclude is None:
    #         exclude = []
    #
    #     errors = {}
    #     for f in self._meta.fields:
    #         if f.name in exclude:
    #             continue
    #         # Skip validation for empty fields with blank=True. The developer
    #         # is responsible for making sure they have a valid value.
    #         raw_value = getattr(self, f.attname)
    #         if f.blank and raw_value in f.empty_values:
    #             continue
    #         try:
    #             if isinstance(f, ForeignKey):
    #             setattr(self, f.attname, f.clean(raw_value, self))
    #         except ValidationError as e:
    #             errors[f.name] = e.error_list
    #
    #     if errors:
    #         raise ValidationError(errors)


from rest_framework.filters import BaseFilterBackend


class GuardianViewPermissionsFilter(BaseFilterBackend):
    """
    A filter backend that limits results to those where they are the foreign key 'user'
    """

    def filter_queryset(self, request, queryset, view):
        return queryset.can_view()


from functools import reduce

from django.db.models import QuerySet
from global_requests import get_thread_user
from guardian.shortcuts import get_objects_for_user
from guardian_queryset.utils import validate_foreign_keys, assign_default_permissions, PUBLIC_VIEW_MODELS_LOWER
from rest_framework.exceptions import ValidationError


class InvalidOperation(Exception):
    pass


class GuardianQuerySet(QuerySet):
    def __init__(self, *args, **kwargs):
        self._reset_validation()
        super().__init__(*args, **kwargs)

    def some_oper_validated(self):
        return reduce(lambda x, value:x or value, self._validated_opers.values(), False)


    def __len__(self):
        if self.some_oper_validated():
            return super().__len__()
        else:
            raise InvalidOperation("No operations have been validated")

    def __iter__(self):
        if not self._validated_opers["view"]:
            return self.can_view().__iter__()
            # raise InvalidOperation("View operations have not been validated")
        return super().__iter__()


    @property
    def model_name(self):
        return self.model.__name__.lower()

    @property
    def user(self):
        return get_thread_user()

    def select_for_update(self, *args, **kwargs):
        obj = super().select_for_update(*args, **kwargs)
        obj.can_change()
        obj.force_validate_oper("view")
        return obj

    def create(self, with_user=True, **kwargs):
        if with_user:
            obj = super().create(**{
                'user': self.user,
                **kwargs,
            })
        else:
            obj = super().create(**kwargs)
        try:
            validate_foreign_keys(self.user, obj)
        except ValidationError as e:
            obj.delete()
            raise e
        assign_default_permissions(self.user, obj)
        return obj


    def _reset_validation(self):
        self._validated_opers = {
            "view": False,
            "change": False,
            "delete": False

        }

    def _clone(self):
        c = super()._clone()
        c._validated_opers = self._validated_opers
        return c


    def can_do(self, perms, user=None):
        user = user or self.user
        if len(perms) == 1 and perms[0] == "view" and self.model_name in PUBLIC_VIEW_MODELS_LOWER:
            self._validated_opers["view"] = True
            return self
        else:
            clone = get_objects_for_user(user, [perm + "_" + self.model_name for perm in perms], klass=self)
            for perm in perms:
                if perm == "view":
                   clone._validated_opers["view"] = True
                elif perm == "change":
                    clone._validated_opers["change"] = True
                elif perm == "delete":
                    clone._validated_opers["delete"] = True
            return clone

    def can_view(self, user=None):
        return self.can_do(["view"], user=user)

    def can_change(self, user=None):
        return self.can_do(["change"], user=user)

    def can_delete(self, user=None):
        return self.can_do(["delete"], user=user)

    def is_owner(self, user=None):
        return self.can_do(["view", "change", "delete"], user=user)

    def force_validate(self):
        for oper in self._validated_opers.keys():
            self._validated_opers[oper] = True
        return self

    def force_validate_oper(self, oper):
        if oper in self._validated_opers.keys():
            self._validated_opers[oper] = True
        return self


    def values(self, *args, **kwargs):
        if not self._validated_opers["view"]:
            raise InvalidOperation("View operations have not been validated")
        return super().values(*args, **kwargs)

    def _values_list(self, *args, **kwargs):
        return super().values_list(*args, **kwargs)
    def values_list(self, *args, **kwargs):
        if not self._validated_opers["view"]:
            raise InvalidOperation("View operations have not been validated")
        return super().values_list(*args, **kwargs)

    def _get(self, *args, **kwargs):
        return super().get(*args, **kwargs)
    def get(self, *args, **kwargs):
        if not self._validated_opers["view"]:
            raise InvalidOperation("View operations have not been validated")
        return super().get(*args, **kwargs)

    def _first(self):
        return super().first()
    def first(self):
        if not self._validated_opers["view"]:
            raise InvalidOperation("View operations have not been validated")
        return super().first()

    def _update(self, **kwargs):
        return super().update(**kwargs)
    def update(self, **kwargs):
        if not self._validated_opers["change"]:
            raise InvalidOperation("Update operations have not been validated")
        return super().update(**kwargs)

    def _delete(self):
        return super().delete()
    def delete(self):
        if not self._validated_opers["delete"]:
            raise InvalidOperation("Delete operations have not been validated")
        return super().delete()

    def _exists(self):
        return super().exists()
    def exists(self):
        if not self.some_oper_validated():
            q = self.can_view()
            self.__dict__.update(q.__dict__)
        return super().exists()




from django.contrib import admin

class GuardianAdminMixin:
    def save_model(self, request, obj, form, change):
        obj._save()

    def get_queryset(self, request):
        return super().get_queryset(request).is_owner()

    def get_field_queryset(self, db, db_field, request):
        """
        If the ModelAdmin specifies ordering, the queryset should respect that
        ordering.  Otherwise don't specify the queryset, let the field decide
        (return None in that case).
        """
        related_admin = self.admin_site._registry.get(db_field.remote_field.model)
        if related_admin is not None:
            ordering = related_admin.get_ordering(request)
            queryset = db_field.remote_field.model._default_manager.using(db)
            if ordering is not None and ordering != ():
                queryset = queryset.order_by(*ordering)
            try:
                return queryset.can_view()
            except:
                return queryset
        return None

    # def get_queryset(self, request):
    #     # use our manager, rather than the default one
    #     qs = self.model._objects.get_queryset()
    #
    #     # we need this from the superclass method
    #     ordering = self.ordering or () # otherwise we might try to *None, which is bad ;)
    #     if ordering:
    #         qs = qs.order_by(*ordering)
    #     return qs

class GuardianModelAdmin(GuardianAdminMixin, admin.ModelAdmin):
    pass


from django_utils.views import as_response
from guardian_queryset.filters import GuardianViewPermissionsFilter



def as_protected_response(
        *args,
        **kwargs
):
    if not "filter_backends" in kwargs:
        kwargs["filter_backends"] = []
    kwargs["filter_backends"].append(GuardianViewPermissionsFilter)
    return as_response(*args, **kwargs)